"""
# Pamda
[![PyPI version](https://badge.fury.io/py/pamda.svg)](https://badge.fury.io/py/pamda)
[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)

Python wrapper for functional programming in object oriented structures.

Inspired heavily by [Ramda](https://ramdajs.com/docs/).


## Documentation for Pamda Functions
https://connor-makowski.github.io/pamda/pamda/pamda.html

## Key Features

- Simplified functional programming for python
- Core Functions include:
  - `curry` arbitrary methods and functions
  - `thunkify` arbitrary methods and functions
  - `pipe` data iteratively through n functions
- List based path access and features for nested dictionaries


## Setup

Make sure you have Python 3.9.x (or higher) installed on your system. You can download it [here](https://www.python.org/downloads/).

### Installation

```
pip install pamda
```

## Getting Started

### Basic Usage
```py
from pamda import pamda

data={'a':{'b':1, 'c':2}}
# Example: Select data given a path and a dictionary
pamda.path(['a','b'])(data) #=> 1

# See documentation for all core pamda functions at
# https://connor-makowski.github.io/pamda/pamda.html
```

### Curry Usage
```py
from pamda import pamda

# Define a function that you want to curry
def myFunction(a,b,c):
    return [a,b,c]

# You can call pamda.curry as a function to curry your functions
curriedMyFn=pamda.curry(myFunction)

# Inputs can now be passed in an async fashion
# The function is evaluated when all inputs are added
x=curriedMyFn(1,2)
x(3) #=> [1,2,3]
x(4) #=> [1,2,4]

# Each set of inputs returns a callable function
# You can stack inputs on a single line for clean functional programming
curriedMyFn(1,2)(3) #=> [1,2,3]
```

For enforcing types, pamda relies on [type_enforced](https://github.com/connor-makowski/type_enforced) but curried objects do not play nice with `type_enforced` objects. To fix this, there is a special curry function, `curryType`, that enables type_enforced annotations for your curried functions:

```py
>>> from pamda import pamda
>>>
>>> # Pamda CurryTyped
>>> @pamda.curryTyped
... def add(a:int,b:int):
...     return a+b
...
>>> add(1)(1)
2
>>> add(1)(1.5)
Traceback (most recent call last):
  File "<stdin>", line 1, in <module>
  File "/home/conmak/development/personal/pamda/pamda/pamda_curry.py", line 43, in __call__
    results = self.__fnExecute__(*new_args, **new_kwargs)
  File "/home/conmak/development/personal/pamda/venv/lib/python3.10/site-packages/type_enforced/enforcer.py", line 90, in __call__
    self.__check_type__(assigned_vars.get(key), value, key)
  File "/home/conmak/development/personal/pamda/venv/lib/python3.10/site-packages/type_enforced/enforcer.py", line 112, in __check_type__
    self.__exception__(
  File "/home/conmak/development/personal/pamda/venv/lib/python3.10/site-packages/type_enforced/enforcer.py", line 34, in __exception__
    raise TypeError(f"({self.__fn__.__qualname__}): {message}")
TypeError: (add): Type mismatch for typed variable `b`. Expected one of the following `[<class 'int'>]` but got `<class 'float'>` instead.
```


### Thunkify Usage
```py
from pamda import pamda

# Define a function that you want to thunkify
# thunkify can be called as a function or decorator
@pamda.thunkify
def myFunction(a,b,c):
    return [a,b,c]

# The function is now curried and the evaluation is lazy
# This means the function is not evaluated until called
x=myFunction(1,2)
x(3) #=> <pamda.curry_obj object at 0x7fd514e4c820>
x(3)() #=> [1,2,3]

y=x(4)
y() #=> [1,2,4]
```

Thunkified functions can be executed asynchronously.

```py
from pamda import pamda
import time

@pamda.thunkify
def test(name, wait):
    print(f'{name} start')
    time.sleep(wait)
    print(f'{name} end')
    return wait

async_test_a = pamda.asyncRun(test('a',2))
async_test_b = pamda.asyncRun(test('b',1))
async_test_a.asyncWait()
async_test_c = pamda.asyncRun(test('c',1))
```

The above code would output:
```
a start
b start
b end
a end
c start
c end
```

### Pipe
```py
from pamda import pamda

def square(x):
    return x**2

def half(x):
    return x/2

def negate(x):
    return -x

# You can pipe data through multiple functions for clean functional programming
pamda.pipe([square, half, negate])(args=(6,),kwargs={}) #=> -18
```

### Use pamda as a subclass
```py
from pamda import pamda

class myClass(pamda):
    def myFunction(self, a):
        return self.inc(a)

mc=myClass()
mc.myFunction(2) #=> 3

@mc.curry
def addUp(a,b):
    return a+b

addUp(1)(2) #=> 3
```

## Pamda Utils

- Pamda also ships with a few helpful utilities
- Check out the documentation here:
  - https://connor-makowski.github.io/pamda/pamda_utils.html"""
from .pamda import pamda



import csv, json
from pamda import pamda_wrappers
import type_enforced


@type_enforced.Enforcer
@pamda_wrappers.staticmethod_wrap
class pamda_utils:
    ######################
    # Data Handling
    def read_csv(
        filename: str,
        return_dict: bool = True,
        cast_items: bool = True,
        cast_dict: dict | str | None = None,
        return_type: str | None = None,
    ):
        """
        Function:

        - Reads the contents of a csv and converts it to list of dicts or list of lists
        - Note: The csv must have a header row indicating the names of each column

        Requires:

        - `filename`:
            - Type: str
            - What: The filepath of the csv to read
                - Note: The first row of the csv must be the header row

        Optional:

        - `return_dict`:
            - Type: bool
            - What: Flag to indicate if the csv should be converted to:
                - True: list of dicts (with each key being the associated column header)
                - False: list of lists (with the first row being the headers)
            - Default: True
            - Notes:
                - This has been deprecated in favor of `return_type`
                - This has been kept for backwards compatibility
                - If return_type is specified, this will be ignored
        - `return_type`:
            - Type: str
            - Options:
                - `list_of_dicts` (default if `return_dict` is True)
                    - A list of dictionaries with each key being the associated column header
                - `dict_of_lists`
                    - A dictionary of lists with each key being the associated column header and each value being a list of the values in that column
                - `list_of_row_lists`
                    - A list of lists (records) with each row being a list of the values in that row
                    - The first row is the header row
                - `list_of_col_lists`
                    - A list of lists (columns) with each column being a list of the values in that column
                    - The first item in each sublist is the header for that column
        - `cast_items`:
            - Type: bool
            - What: Flag to indicate if an attempt to cast each item to a proper type
            - Default: True
            - Note: This is useful for converting strings to ints, floats, etc.
            - Note: This works in conjunction with `cast_dict`
                - If `cast_dict` is not None, then an automated attempt to cast the items will be made
                - For automated casting, the following rules are applied to each item in the data:
                    - If the item is a string:
                        - If the string is empty, `None` will be returned
                        - If the string is "None" or "null", `None` will be returned
                        - If the string is "True" or "true", `True` will be returned
                        - If the string is "False" or "false", `False` will be returned
                        - If the string is a valid float, the float will be returned
                        - If the string is a valid int, the int will be returned
                        - Otherwise, the string will be returned
                    - If the item is not a string, it will be returned as is
        - `cast_dict`:
            - Type: dict
            - What: A dictionary of functions to cast each column (by name) in the csv
            - Default: None
            - Note: Unspecified column names will be treated as strings
            - Note: `cast_items` must be `True` to use this
            - EG: {
                'user_id': lambda x: int(x),
                'year': lambda x: int(x),
                'pass': lambda x: x.lower()=='true',
            }
        """
        assert return_type in [
            None,
            "list_of_dicts",
            "dict_of_lists",
            "list_of_row_lists",
            "list_of_col_lists",
        ], f"Invalid return_type: {return_type}"
        with open(filename) as f:
            file_data = csv.reader(f, delimiter=",", quotechar='"')
            headers = next(file_data)
            data = list(zip(*[row for row in file_data]))
        if cast_items:
            if cast_dict is not None:
                for idx, header in enumerate(headers):
                    cast_fn = cast_dict.get(header, lambda x: x)
                    data[idx] = [cast_fn(item) for item in data[idx]]
            else:

                def cast(obj):
                    if not isinstance(obj, str):
                        return obj
                    obj_lower = obj.lower()
                    if obj == "" or obj_lower == "none" or obj_lower == "null":
                        return None
                    if obj_lower == "true":
                        return True
                    if obj_lower == "false":
                        return False
                    try:
                        float_obj = float(obj)
                        return (
                            int(float_obj)
                            if float_obj == int(float_obj)
                            else float_obj
                        )
                    except:
                        return obj

                for idx, header in enumerate(headers):
                    data[idx] = [cast(item) for item in data[idx]]
        # Maintain backwards compatibility
        # TODO: Deprecate this in the next major release
        if return_type == None:
            return_type = (
                "list_of_dicts" if return_dict else "list_of_row_lists"
            )
        if return_type == "list_of_dicts":
            return [dict(zip(headers, row)) for row in zip(*data)]
        elif return_type == "dict_of_lists":
            return {header: col for header, col in zip(headers, data)}
        elif return_type == "list_of_row_lists":
            return [headers] + [list(row) for row in zip(*data)]
        elif return_type == "list_of_col_lists":
            return [[header] + list(col) for header, col in zip(headers, data)]

    def write_csv(filename: str, data):
        """
        Function:

        - Writes the contents of a list of list or list of dicts to a csv

        Requires:

        - `filename`:
            - Type: str
            - What: The filepath of the csv to read
        - `data`:
            - Type: list of lists | list of dicts
            - What: The data to write
        """
        with open(filename, "w") as f:
            if isinstance(data[0], dict):
                writer = csv.DictWriter(f, fieldnames=data[0].keys())
                writer.writeheader()
                for row in data:
                    writer.writerow(row)
            elif isinstance(data[0], list):
                writer = csv.writer(f)
                for row in data:
                    writer.writerow(row)
            else:
                raise Exception(
                    "`write_csv` takes in list of lists or list of dicts only."
                )

    def read_json(filename: str):
        """
        Function:

        - Reads the contents of a json

        Requires:
        - `filename`:
            - Type: str
            - What: The filepath of the json to read
        """
        const_map = {
            "-Infinity": float("-Infinity"),
            "Infinity": float("Infinity"),
            "NaN": None,
        }
        with open(filename) as f:
            return json.load(f, parse_constant=lambda x: const_map[x])

    def write_json(filename: str, data, pretty: bool = False):
        """
        Function:

        - Writes the contents of a list of list or list of dicts to a json

        Requires:

        - `filename`:
            - Type: str
            - What: The filepath of the json to write
        - `data`:
            - Type: A json serializable python object
            - What: The data to write
        """
        with open(filename, "w") as f:
            if pretty:
                json.dump(data, f, indent=4)
            else:
                json.dump(data, f)

    ######################
    # Helpful Functions
    def getMethods(object):
        """
        Function:

        - Returns the callable methods of a class (dunder-excluded) as a list of strs

        Requires:

        - `object`:
            - Type: any
            - What: Any python object
            - Default: 1

        Example:

        ```
        class MyClass:
            def A(self):
                pass

            def B(self):
                pass


        pamda.getMethods(MyClass) #=> ['A', 'B']
        ```
        """
        return [
            fn
            for fn in dir(object)
            if callable(getattr(object, fn)) and not fn.startswith("__")
        ]

    def getForceDict(object: dict | list, key: str | int):
        """
        Function:

        - Returns a value from a dictionary (or list) given a key (or index)  and forces that value to be a dictionary if it is not a dictionary (or a list)
        - Note: This updates the object in place to force the value from the key to be a dictionary

        Requires:

        - `object`:
            - Type: dict | list
            - What: The object from which to look for a key or index
        - `key`:
            - Type: str | int
            - What: The key or index to look up in the object

        Example:

        ```
        data = {'a':{}, 'b':1, 'c':[]}

        pamda.getForceDict(data, 'a') #=> {}
        pamda.getForceDict(data, 'b') #=> {}
        pamda.getForceDict(data, 'c') #=> []

        # Note that the object has been updated in place
        data #=> {'a':{}, 'b':{}, 'c':[]}
        ```
        """
        # TODO: Clean this up for faster performance
        if isinstance(object, dict):
            value = object.get(key)
        else:
            value = object[key]
        if not isinstance(value, (dict, list)):
            value = {}
            object.__setitem__(key, value)
        return value



import time


class pamda_timer:
    def __init__(self, __fn__):
        """
        Function:

        Initialize a pamda_timer object.
        - Note: This object is used for bare minimum in script timing purposes.

        Required:

        - `fn`:
            - Type: function | method
            - What: The name of the process being timed
        """
        self.__fn__ = __fn__

    def __call__(self, *args, **kwargs):
        """
        Function:

        Test the function processing time.

        Optional:

        - `args`:
            - Type: any
            - What: The arguments to pass to the function
        - `kwargs`:
            - Type: any
            - What: The keyword arguments to pass to the function
        """
        start = time.time()
        out = self.__fn__(*args, **kwargs)
        print(f"{self.__fn__.__qualname__}: {round(time.time() - start,4)}s")
        return out

    def __repr__(self):
        return f"<timed {self.__fn__.__module__}.{self.__fn__.__qualname__} object at {hex(id(self))}>"
